#version 140

#define MAX_LIGHTS 8

#define LIGHT_OMNI 0
#define LIGHT_OMNINORMAL 1
#define LIGHT_SPOT 2
#define LIGHT_SPOTNORMAL 3
#define LIGHT_SHADOW 4

#define MAX_EFFECTS 4

#define EFFECT_NONE 0
#define EFFECT_FADE 1
#define EFFECT_GHOST 2
#define EFFECT_SPLASH 3
#define EFFECT_CHROME 4
#define EFFECT_BANG 5
#define EFFECT_ELECTRO 6
#define EFFECT_TURBO 7
#define EFFECT_GLOW 8
#define EFFECT_FOX 9
#define EFFECT_WATERBOX 10
#define EFFECT_SHOCKWAVE 11
#define EFFECT_BOMB 12

in vec4 inPosition;
in vec4 inColor;
in vec2 inTexCoord;
in float inFogCoord;
in vec3 inNormal;

out vec4 varColor;
out vec2 varTexCoord;
out vec2 varEnvCoord;
out float varFog;

// projection matrix
uniform mat4 matViewProj;
uniform vec4 modelUpVec;

// fog vars
uniform bool inFog;
uniform vec2 fogParams;
uniform bool isModelMirror;
uniform vec3 mirrorParams;

// env vars
uniform bool setEnv;
uniform vec3 cameraPos;
uniform vec3 cameraUpVec;
uniform bool isModel;
uniform bool modelEnvGood;
uniform vec3 modelEnvMatX;
uniform vec3 modelEnvMatY;

// lights
uniform int numLights;
uniform vec3 lightPos[MAX_LIGHTS];
uniform vec3 lightColor[MAX_LIGHTS];
uniform vec3 lightParams[MAX_LIGHTS];
uniform mat3 lightDir[MAX_LIGHTS];
uniform bool modelShadow;
uniform vec3 shadowColor;
uniform bool addLight;
uniform vec3 modelLight;

// special fx
uniform int modelEffect;
uniform vec2 modelFxParams;
uniform int numEffects;
uniform int effectType[MAX_EFFECTS];
uniform vec3 effectParams[MAX_EFFECTS];
uniform vec3 effectParams2[MAX_EFFECTS];
uniform float wboxSineTable[9];

bool inShadow;

void ProcessShadow(vec3 lightPos, vec3 lightSize, mat3 lightDir)
{
  if (!inShadow) {
    vec3 LightDelta = lightPos - inPosition.xyz;

    float dist = dot(LightDelta, lightDir[0]);
    if (dist < -lightSize.x || dist > lightSize.x) {
      return;
    }

    dist = dot(LightDelta, lightDir[1]);
    if (dist < -lightSize.y || dist > lightSize.y) {
      return;
    }

    dist = dot(LightDelta, lightDir[2]);
    if (dist < -lightSize.z || dist > lightSize.z) {
      return;
    }

    inShadow = true;
  }
}

vec4 ProcessLight(vec3 lightPos, vec3 lightColor, vec3 lightParams, mat3 lightDir)
{
  vec4 col = vec4(0.0, 0.0, 0.0, 0.0);
  int type = int(lightParams.x);

  // special case for shadow box
  if (type == LIGHT_SHADOW) {
    ProcessShadow(lightPos, lightColor, lightDir);
    return col;
  }

  float reach = lightParams.y;
  vec3 LightDelta = lightPos - inPosition.xyz;
  float dist = dot(LightDelta, LightDelta);
  if (dist >= reach) {
    return col;
  }

  float scale = 1.0 - (dist / reach);

  if (type != LIGHT_OMNI) {
    dist = sqrt(dist);
  }

  // calc angle from normal if needed
  if (type == LIGHT_OMNINORMAL || type == LIGHT_SPOTNORMAL) {
    float ang = dot(LightDelta, inNormal);
    if (ang <= 0.0) {
      return col;
    }
    ang = clamp(ang / dist, 0.0, 1.0);
    scale *= ang;
  }

  // calc cone adjustment if needed
  if (type == LIGHT_SPOT || type == LIGHT_SPOTNORMAL) {
    float conemul = lightParams.z;
    float cone = -dot(lightDir[2], LightDelta) / dist - 1.0;
    cone = cone * conemul + 1.0;
    if (cone <= 0.0) {
      return col;
    }
    cone = clamp(cone, 0.0, 1.0);
    scale *= cone;
  }

  col.rgb = mix(col.rgb, lightColor, scale);
  return col;
}

void ProcessMeshEffect(inout vec4 pos, int effectType, vec3 effectParams, vec3 effectParams2)
{
  // waterbox mesh fx
  if (effectType == EFFECT_WATERBOX) {
    vec3 bboxMin = effectParams.xyz;
    vec3 bboxMax = effectParams2.xyz;
    if (all(greaterThan(pos.xyz, bboxMin)) && all(lessThan(pos.xyz, bboxMax))) {
      int hash = int(abs(pos.x * pos.z));
      pos.x += wboxSineTable[(hash + 0) % 9];
      pos.y += wboxSineTable[(hash + 3) % 9] + 6.0;
      pos.z += wboxSineTable[(hash + 6) % 9];
    }

  // shockwave mesh fx
  } else if (effectType == EFFECT_SHOCKWAVE) {
    vec3 objPos = effectParams.xyz;
    float reach = effectParams2.x;

    vec3 delta = pos.xyz - objPos;
    float dist = dot(delta, delta);
    if (dist < reach * reach) {
      dist = sqrt(dist);
      float pull = (reach - dist) * 0.1;
      pull = min(pull, dist * 0.5);
      pos.xyz -= delta * (pull / dist);
    }

  // putty bomb mesh fx
  } else if (effectType == EFFECT_BOMB) {
    vec3 objPos = effectParams.xyz;
    float rad = effectParams2.x;
    float timeStep = effectParams2.y;

    vec3 delta = pos.xyz - objPos;
    float dist = dot(delta, delta);
    dist = sqrt(dist);

    float scale = (0.5 - timeStep) / 0.5;
    float push = (64.0 - abs(rad - dist)) * scale;
    if (push > 0.0) {
      pos.xyz += delta * (push / dist);
    }
  }
}

void ProcessWeaponEffect(inout vec4 pos)
{
  float timeStep = modelFxParams.x;
  float time = inTexCoord.s;
  float add = inTexCoord.t;
  time += add * timeStep;

  // bomb bang
  if (modelEffect == EFFECT_BANG) {
    varTexCoord.s = sin(time) * (12.0f / 256.0f) + (36.0f / 256.0f);
    varTexCoord.t = cos(time) * (12.0f / 256.0f) + (68.0f / 256.0f);

  // electro pulse
  } else if (modelEffect == EFFECT_ELECTRO) {
    varTexCoord.s = sin(time) * (24.0f / 256.0f) + (96.0f / 256.0f);
    varTexCoord.t = cos(time) * (24.0f / 256.0f) + (96.0f / 256.0f);

  // turbo
  } else if (modelEffect == EFFECT_TURBO) {
    varTexCoord.s = sin(time) * (10.0f / 256.0f) + (96.0f / 256.0f);
    varTexCoord.t = cos(time) * (10.0f / 256.0f) + (160.0f / 256.0f);

  // bomb glow
  } else if (modelEffect == EFFECT_GLOW) {
    varTexCoord.s = sin(time) * (12.0f / 256.0f) + (36.0f / 256.0f);
    varTexCoord.t = cos(time) * (12.0f / 256.0f) + (44.0f / 256.0f);

  // fox
  } else if (modelEffect == EFFECT_FOX) {
    varTexCoord.s = sin(time) * (24.0f / 256.0f) + (160.0f / 256.0f);
    varTexCoord.t = cos(time) * (24.0f / 256.0f) + (160.0f / 256.0f);
  }
}

void ProcessModelEffect(inout vec4 pos)
{
  // fade alpha
  if (modelEffect == EFFECT_FADE) {
    float timeStep = modelFxParams.x;
    varColor.a = 1.0 - timeStep;

  // ghost car
  } else if (modelEffect == EFFECT_GHOST) {
    float ghostPos = modelFxParams.x;
    float ghostMul = modelFxParams.y;
    varColor.a = abs(pos.z + ghostPos) * ghostMul;
    varColor.a = 1.0 - varColor.a;

  // water splash
  } else if (modelEffect == EFFECT_SPLASH) {
    float timeStep = modelFxParams.x;
    float frame = inPosition.w * timeStep;
    if (frame < 16.0) {
      float grav = 384.0 * timeStep;
      vec3 vel = inNormal;
      vel.y += grav;
      pos.xyz += vel * timeStep;
      pos.w = 1.0;

      float rgb = (16.0 - frame) / 32.0;
      varColor.rgb = vec3(rgb);
      varTexCoord.s += floor(frame) / 16.0;
    }

  // chrome ball
  } else if (modelEffect == EFFECT_CHROME) {
    varTexCoord.s = dot(inNormal, modelEnvMatX) + 0.5;
    varTexCoord.t = dot(inNormal, modelEnvMatY) + 0.6;

  // weapon fx
  } else {
    ProcessWeaponEffect(pos);
    return;
  }
}

void main(void)
{
  vec4 pos = inPosition;

  varColor = inColor;
  varTexCoord = inTexCoord;

  // process effects
  if (modelEffect != EFFECT_NONE) {
    ProcessModelEffect(pos);
  }

  if (numEffects > 0) {
    ProcessMeshEffect(pos, effectType[0], effectParams[0], effectParams2[0]);
  }
  if (numEffects > 1) {
    ProcessMeshEffect(pos, effectType[1], effectParams[1], effectParams2[1]);
  }
  if (numEffects > 2) {
    ProcessMeshEffect(pos, effectType[2], effectParams[2], effectParams2[2]);
  }
  if (numEffects > 3) {
    ProcessMeshEffect(pos, effectType[3], effectParams[3], effectParams2[3]);
  }

  // calc position
  gl_Position = pos * matViewProj;

  // is it in fog?
  if (inFog) {
    float fogStart = fogParams.x;
    float fogEnd = fogParams.y;
    float fogDist = fogEnd - fogStart;

    float vertFog;
    if (isModelMirror) {
      float mirrorHeight = mirrorParams.x;
      float mirrorMul = mirrorParams.y;
      float mirrorAdd = mirrorParams.z;
      vertFog = dot(pos, modelUpVec);
      vertFog = mirrorHeight - vertFog;
      vertFog *= mirrorMul;
      vertFog += mirrorAdd;
      clamp(vertFog, 0.0, 1.0);
    } else {
      vertFog = inFogCoord;
    }

    varFog = max(gl_Position.w, fogStart);
    varFog += vertFog * fogDist;
    varFog = (fogEnd - varFog) / fogDist;
    varFog = clamp(varFog, 0.0, 1.0);
  }

  // calc env tex coords
  if (setEnv) {
    if (isModel && !modelEnvGood) {
      varEnvCoord.s = dot(inNormal, modelEnvMatX) + 0.5;
      varEnvCoord.t = dot(inNormal, modelEnvMatY) + 0.5;
    } else {
      vec3 vecz = normalize(pos.xyz - cameraPos);
      vec3 vecx = cross(cameraUpVec, vecz);
      vec3 vecy = cross(vecz, vecx);
      varEnvCoord.s = dot(inNormal, vecx) * 0.5 + 0.5;
      varEnvCoord.t = dot(inNormal, vecy) * 0.5 + 0.5;
    }
  }

  // process lights
  inShadow = modelShadow;

  if (numLights > 0) {
    varColor += ProcessLight(lightPos[0], lightColor[0], lightParams[0], lightDir[0]);
  }
  if (numLights > 1) {
    varColor += ProcessLight(lightPos[1], lightColor[1], lightParams[1], lightDir[1]);
  }
  if (numLights > 2) {
    varColor += ProcessLight(lightPos[2], lightColor[2], lightParams[2], lightDir[2]);
  }
  if (numLights > 3) {
    varColor += ProcessLight(lightPos[3], lightColor[3], lightParams[3], lightDir[3]);
  }
  if (numLights > 4) {
    varColor += ProcessLight(lightPos[4], lightColor[4], lightParams[4], lightDir[4]);
  }
  if (numLights > 5) {
    varColor += ProcessLight(lightPos[5], lightColor[5], lightParams[5], lightDir[5]);
  }
  if (numLights > 6) {
    varColor += ProcessLight(lightPos[6], lightColor[6], lightParams[6], lightDir[6]);
  }
  if (numLights > 7) {
    varColor += ProcessLight(lightPos[7], lightColor[7], lightParams[7], lightDir[7]);
  }

  if (inShadow) {
    varColor += vec4(shadowColor, 0.0);
  }

  // add any tinting
  if (addLight) {
    varColor += vec4(modelLight, 0.0);
  }

  // clamp the color
  varColor = clamp(varColor, 0.0, 1.0);
}
